"""
Defines the views for the API
"""

# from django.shortcuts import render
from rest_framework import viewsets
from rest_framework.views import APIView
from rest_framework.response import Response
from django.shortcuts import get_object_or_404
from django.http import HttpResponse
from .pagination import PageNumberWithPageSizePagination
import copy

from .course_scraper.course_scraper import scraping

from .serializers import (
    SmallSportListSerializer,
    SportListSerializer,
    CriterionListSerializer,
    QuestionListSerializer,
    SingleSportSerializer,
    CriteriaSerializer,
    IncompleteSportSerializer,
    ArchiveSerializer,
)
from .models import Sport, Criterion, Question

# Create your views here.


class SportListView(viewsets.ModelViewSet):  # pylint: disable=too-many-ancestors
    """
    A View returning every Sport Object
    """

    serializer_class = SportListSerializer
    queryset = Sport.objects.all()


class CriterionListView(viewsets.ModelViewSet):  # pylint: disable=too-many-ancestors
    """
    A View returning every Criterion Object
    """

    serializer_class = CriterionListSerializer
    queryset = Criterion.objects.all()


class QuestionListView(viewsets.ModelViewSet):  # pylint: disable=too-many-ancestors
    """
    A View returning every Question Object
    """

    serializer_class = QuestionListSerializer
    queryset = Question.objects.all()


# Dev Notes:
# - If we want to include a View in the Router in urls.py, the View needs to be a Viewset
# - Those are mostly meant for Lists of Objects, so instead of get() and post(), list() and create() are used respectively
# https://stackoverflow.com/questions/30389248/how-can-i-register-a-single-view-not-a-viewset-on-my-router


class SmallSportListView(viewsets.ViewSet):
    """
    View for Sports and List of Sport:
    List returns every Sport with the is_filled Field
    Detail returns single Sports with every Criterium
    """

    authentication_classes = []

    # GET for api/admin/sport/
    def list(self, request):
        """
        GET for api/admin/sport/
        Returns a List of Every Sport with the is_filled Field, stating whether every Criterion is given a Rating or not
        """

        paginator = PageNumberWithPageSizePagination()

        sports = Sport.objects.filter(currently_active=True).order_by("name")
        sports = paginator.paginate_queryset(sports, request)
        is_filled_tuples = []

        for sport in sports:

            is_filled_tuples.append((sport, sport.is_filled()))

        serializer = SmallSportListSerializer(is_filled_tuples)

        return paginator.get_paginated_response(serializer.data)

    # POST for api/admin/sport/
    def create(self, request):
        """
        POST for api/admin/sport/
        View for Creating a New Sport
        """

        request_data = SingleSportSerializer().to_internal_value(request)

        # A Try at Error Catching
        if request_data is None:
            return HttpResponse(status=400)

        new_sport = Sport.objects.create_sport()

        new_sport.name = request_data["name"]
        new_sport.url = request_data["url"]

        new_sport.save()

        response = SingleSportSerializer(new_sport)

        return Response(response.data)

    # GET for api/admin/sport/<id>/
    def retrieve(self, request, pk=None):
        """
        GET for api/admin/sport/<pk>/
        View for getting a Single Sport, with the pk to the Sport being the argument in the URL
        """

        sport = get_object_or_404(Sport, pk=pk)

        response = SingleSportSerializer(sport)

        return Response(response.data)

    # PUT for api/admin/sport/<id>/
    def update(self, request, pk=None):
        """
        PUT for api/admin/sport/<id>/
        Creates a Sport if it doesn't exist, otherwise overwrites it.
        TODO: Maybe Rework PUT if needed of Admin Frontend
        """

        # Get Data from Serializer
        request_data = SingleSportSerializer().to_internal_value(request)

        if request_data is None:
            # Something is Broke, so Refuse Changing Data
            return HttpResponse(status=400)

        # Get Sport from Data
        sport = Sport.objects.get(pk=pk)

        # Apply New Data to Sport
        sport.name = request_data["name"]
        sport.url = request_data["url"]

        if request_data["currently_active"] is True and sport.currently_active is False:
            # If the Sport is just now being activated, update the Date
            sport.reactivate()
        else:
            sport.currently_active = request_data["currently_active"]

        # Overwrite Criterion Ratings
        for criterion, value in request_data["criteria"]:
            sport.rate(criterion, value)

        sport.save()

        # Re-Serialize changed Sport and Send it back
        response = SingleSportSerializer(sport)

        return Response(response.data)

    # PATCH for api/admin/sport/<id>/
    def partial_update(self, request, pk=None):
        """
        PATCH for api/admin/sport/<id>/
        Fills in the given Values into the Sport specified in the URL
        """

        # Get Data from Serializer
        request_data = SingleSportSerializer().to_internal_value(request)

        if request_data is None:
            # Something is Broke, so Refuse Changing Data
            return HttpResponse(status=400)

        # Get Sport from Data
        sport = Sport.objects.get(pk=pk)

        # Apply New Data to Sport, if it exists
        if "name" in request_data.keys():
            sport.name = request_data["name"]

        if "url" in request_data.keys():
            sport.url = request_data["url"]

        if "currently_active" in request_data.keys():
            if (
                request_data["currently_active"] is True
                and sport.currently_active is False
            ):
                # If the Sport is just now being activated, update the Date
                sport.reactivate()
            else:
                sport.currently_active = request_data["currently_active"]

        # Overwrite Criterion Ratings
        if "criteria" in request_data.keys():
            for criterion, value in request_data["criteria"]:
                sport.rate(criterion, value)

        sport.save()

        # Re-Serialize changed Sport and Send it back
        response = SingleSportSerializer(sport)

        return Response(response.data)

    # DELETE for api/admin/sport/<id>/
    def destroy(self, request, pk=None):
        """
        DELETE for api/admin/sport/<id>/
        Removes Sport Object specified in the URL
        """

        sport = get_object_or_404(Sport, pk=pk)

        sport.delete()

        return HttpResponse(status=404)


class IncompleteSportView(APIView):
    """
    Returns all Sport Objects with Incomplete Ratings
    """

    authentication_classes = []

    # GET for api/admin/sport/incomplete/
    def get(self, request):
        """
        GET for api/admin/sport/incomplete/
        Returns every incomplete Sport with its incomplete Ratings in a paginated manner
        """

        paginator = PageNumberWithPageSizePagination()
        queryset = Sport.objects.all().order_by("name")
        queryset = paginator.paginate_queryset(queryset, request)

        incomplete_sport_list = []

        for sport in Sport.objects.filter(currently_active=True).iterator():

            if not sport.is_filled():
                incomplete_sport_list.append(sport)

        response = IncompleteSportSerializer(incomplete_sport_list)

        return paginator.get_paginated_response(response.data)


class CriteriaView(APIView):
    """
    View for the List of Criteria and their Metadata
    """

    authentication_classes = []

    # GET for api/admin/criteria/
    def get(self, request):
        """
        GET for api/admin/criteria/
        Returns every Criterium and the Metadata of
        Number of Sports in which the Rating is >1
        and the cumulated sum of Ratings >1
        TODO: Also Pagination
        """

        data = []

        for crit in Criterion.objects.iterator():

            active_sports, sum_of_weights = crit.get_active_sum()

            data.append((crit, active_sports, sum_of_weights))

        response = CriteriaSerializer(data)

        return Response(response.data)


class SportArchiveView(APIView):
    """
    View For the List of all Archived Sports (so all sports with currently_active=False)
    """

    def get(self, request):
        """
        GET for api/admin/archive/
        """

        paginator = PageNumberWithPageSizePagination()

        archived_sports = paginator.paginate_queryset(
            Sport.objects.filter(currently_active=False).order_by("name"), request
        )

        response = ArchiveSerializer(archived_sports)

        return paginator.get_paginated_response(response.data)


class ScraperView(APIView):
    """
    View for the Scraper, including GET and POST
    """

    def get(self, request):

        # Scrape Sports from their website
        scraped_sports = scraping()

        # Reformat scraped Sports so they have a name and an url field
        new_sport_list = []
        for sport_name, sport_url in scraped_sports.items():
            new_sport_list.append({"name": sport_name, "url": sport_url})

        # Iterate through DB Sport entries to see which ones are in the new Sport list etc
        diff_list = []

        id_counter = 1

        empty_diff_dict = {
            "id": -1,
            "kind_of_diff": "",
            "old_sport": {
                "id": -1,
                "name": "",
                "url": "",
                "last_used": "",
            },
            "new_sport": {
                "name": "",
                "url": "",
            },
        }

        for old_sport in Sport.objects.filter(currently_active=True):

            diff_dict = copy.deepcopy(empty_diff_dict)

            diff_dict["id"] = id_counter

            diff_dict["old_sport"]["id"] = old_sport.pk
            diff_dict["old_sport"]["name"] = old_sport.name
            diff_dict["old_sport"]["url"] = old_sport.url
            diff_dict["old_sport"]["last_used"] = old_sport.last_used

            if old_sport.name in scraped_sports.keys():
                # A Sport currently active is also found in the new scraped sports
                # -> "same" Sport

                diff_dict["kind_of_diff"] = "same"

                new_sport_url = scraped_sports.pop(old_sport.name)

                diff_dict["new_sport"]["name"] = old_sport.name
                diff_dict["new_sport"]["url"] = new_sport_url

            else:
                # A Sport currently active is _not_ found in the new scraped sports
                # -> "to_be_archived" Sport

                diff_dict["kind_of_diff"] = "to_be_archived"

            diff_list.append(diff_dict)
            id_counter += 1

        for new_sport_name, new_sport_url in scraped_sports.items():

            # The query should only contain zero values, or one value.
            # The Name is technically no PK-Candidate, but they shouldn't be included more than once
            query = Sport.objects.filter(name=new_sport_name)

            diff_dict = copy.deepcopy(empty_diff_dict)

            diff_dict["id"] = id_counter

            diff_dict["new_sport"]["name"] = new_sport_name
            diff_dict["new_sport"]["url"] = new_sport_url

            if query.count() == 0:
                # The new Sport is not found in the Archive, so it is completely new
                # -> "new" Sport

                diff_dict["kind_of_diff"] = "new"

            else:
                # The new Sport is in the Archive, so it needs to be resurrected
                # -> "from_archive" Sport

                old_sport = query.get(name=new_sport_name)

                diff_dict["kind_of_diff"] = "from_archive"

                diff_dict["old_sport"]["id"] = old_sport.pk
                diff_dict["old_sport"]["name"] = old_sport.name
                diff_dict["old_sport"]["url"] = old_sport.url
                diff_dict["old_sport"]["last_used"] = old_sport.last_used

            diff_list.append(diff_dict)
            id_counter += 1

        # Paginate it all!

        paginator = PageNumberWithPageSizePagination()
        paginated_list = paginator.paginate_queryset(diff_list, request)

        return paginator.get_paginated_response(paginated_list)

    def post(self, request):
        """
        Gets List of Diffs from Frontend and writes them into the Database
        """

        diff_data = request.data

        for diff in diff_data:

            if diff["kind_of_diff"] == "new":

                sport = Sport.objects.create_sport()
                sport.name = diff["new_sport"]["name"]
                sport.url = diff["new_sport"]["url"]

            elif diff["kind_of_diff"] == "same":
                # In Case of URL changes, "same" updates the URL

                sport = Sport.objects.get(pk=diff["old_sport"]["id"])

                sport.url = diff["new_sport"]["url"]

                # To Update last_used
                sport.reactivate()

            elif diff["kind_of_diff"] == "to_be_archived":

                sport = Sport.objects.get(pk=diff["old_sport"]["id"])
                sport.currently_active = False

            else:
                # -> from_archive

                sport = Sport.objects.get(pk=diff["old_sport"]["id"])
                sport.reactivate()

                # Same as in "same", url could be different
                sport.url = diff["new_sport"]["url"]

            sport.save()

        # TODO: Maybe Redirect?
        return Response(status=200)
